/* Copyright (C) 2012,2013, Matthieu Nottale
* See README.md for license
*/

#include <boost/asio.hpp>
#include <boost/thread.hpp>
#include <boost/chrono.hpp>
#include <boost/function.hpp>
#include <boost/utility.hpp>

#include "libclient.hh"

static bool debugOn = getenv("DEBUG");

#define DEBUG(a) do { if (debugOn) std::cerr << a << std::endl;} while(false)
namespace ptrim
{

DeviceColor::DeviceColor()
{
  brightness = 255;
  memset(colors, 0, sizeof(colors));
}

DeviceColor::DeviceColor(const std::string& idata, unsigned char brightness)
{
  // fixme: extract this code, generalize to N leds
  // remove all spaces
  std::string data;
  for (unsigned i=0; i<idata.size(); ++i)
    if (idata[i] != ' ')
      data += idata[i];
  this->brightness = brightness;
  switch(data.size())
  {
  case 1:
    {
      unsigned char c = strtol(data.c_str(), 0, 16) * 16;
      set(c);
      break;
    }
  case 2:
    { // one intensity for all leds
      unsigned char c = strtol(data.c_str(), 0, 16);
      set(c);
      break;
    }
  case 3:
    { // one 1-byte RGB for all leds
      RGB c;
      c.v[0] = strtol(data.substr(0, 1).c_str(), 0, 16)*16;
      c.v[1] = strtol(data.substr(1, 1).c_str(), 0, 16)*16;
      c.v[2] = strtol(data.substr(2, 1).c_str(), 0, 16)*16;
      set(c);
      break;
    }
  case 6:
    { // one 2-byte RGB for all leds
      RGB c;
      c.v[0] = strtol(data.substr(0, 2).c_str(), 0, 16);
      c.v[1] = strtol(data.substr(2, 2).c_str(), 0, 16);
      c.v[2] = strtol(data.substr(4, 2).c_str(), 0, 16);
      set(c);
      break;
    }
  case 24:
    { // one-byte RGB per led
      for (unsigned l=0; l<24; l+=3)
      {
        RGB c;
        c.v[0] = strtol(data.substr(l+0, 1).c_str(), 0, 16)*16;
        c.v[1] = strtol(data.substr(l+1, 1).c_str(), 0, 16)*16;
        c.v[2] = strtol(data.substr(l+2, 1).c_str(), 0, 16)*16;
        set(l/3, c);
      }
      break;
    }
  case 48:
    {
      for (unsigned l=0; l<48; l+=6)
      {
        RGB c;
        c.v[0] = strtol(data.substr(l+0, 2).c_str(), 0, 16);
        c.v[1] = strtol(data.substr(l+2, 2).c_str(), 0, 16);
        c.v[2] = strtol(data.substr(l+4, 2).c_str(), 0, 16);
        set(l/6, c);
      }
      break;
    }
  default:
    throw std::runtime_error("Invalid color string size");
  }
}

void DeviceColor::set(unsigned char c)
{
  memset(colors, c, 24);
}

void DeviceColor::set(RGB c)
{
  for (unsigned i=0; i<8; ++i)
    colors[i] = c;
}

void DeviceColor::set(unsigned p, RGB c)
{
  colors[p] = c;
}

class Char
{
public:
  Char(unsigned char c)
  : c(c)
  {
  }
  unsigned char c;
};

std::ostream& operator << (std::ostream& o, const Char& c)
{
  if (c.c < 16)
    o << '0';
  o << (unsigned int)c.c;
  return o;
}

std::string DeviceColor::toString() const
{
  std::stringstream s;
  std::hex(s);
  for (unsigned i=0; i<8; ++i)
    s << Char(colors[i].r)
      << Char(colors[i].g)
      << Char(colors[i].b);
  return s.str();
}

using namespace boost::asio;

class NetClientImpl
{
public:
  NetClientImpl(): s(io) {}
  io_service io;
  ip::tcp::socket s;
  std::vector<std::string> devices;
  char buffer[4096];
};

NetClient::NetClient()
{
  _p = new NetClientImpl();
}

NetClient::~NetClient()
{
  delete _p;
}

const std::vector<std::string>& NetClient::devices()
{
  return _p->devices;
}

void NetClient::connect(const std::string& host, const std::string& port)
{
  ip::tcp::resolver resolver(_p->io);
  ip::tcp::resolver::query query(host, port);
  ip::tcp::resolver::iterator endpoint_iterator = resolver.resolve(query);

  boost::asio::connect(_p->s, endpoint_iterator);
  std::string devices = "devices\n";
  write(_p->s, buffer(devices.c_str(), devices.length()));

  // read until we get the ready, getting device names in the process
  streambuf response;
  while(true)
  {
    read_until(_p->s, response, "\n");
    std::istream response_stream(&response);
    std::string line;
    std::getline(response_stream, line);
    if (line.empty())
      break;
    std::cerr << "got '" << line<< std::endl;
    _p->devices.push_back(line);
  }
  std::cerr << "finished" << std::endl;
}

void NetClient::setColor(const std::string& device, const DeviceColor& color)
{
  std::stringstream ss;
  ss << "color " << color.toString() << " " << 255 << " " << device << "\n";

  std::string st = ss.str();
  write(_p->s, buffer(st.c_str(), st.length()));
}


Group::Group(Client& client, const std::vector<std::string> devices)
: client(client)
, devices(devices)
{
  colors.resize(devices.size());
}

void Group::setColor(const std::vector<DeviceColor>& data, const Mask& m)
{
  DEBUG("Group::set " << data.size());
  for (unsigned i = 0; i< data.size(); ++i)
  {
    unsigned char mask = m.mask[i];
    for (unsigned j=0; j<8; ++j)
      if (mask & (1<<j))
        colors[i].colors[j] = data[i].colors[j];
    if (mask)
      client.setColor(devices[i], colors[i]);
  }
}

Region::Region(Group& g, unsigned firstDev, unsigned lastDev, unsigned firstLed, unsigned lastLed)
: group(g)
, firstDev(firstDev), lastDev(lastDev), firstLed(firstLed), lastLed(lastLed)
{
  if (lastDev >= g.devices.size() || lastLed >=8)
    throw std::runtime_error("Region out of range");
}


void Region::setColor(const unsigned char* data)
{
  for (unsigned d=firstDev; d<=lastDev; ++d)
  {
    DeviceColor& col = group.colors[d];
    for (unsigned l = firstLed; l<=lastLed; ++l)
    {
      memcpy(col.colors[l].v, data, 3);
      data += 3;
    }
    group.client.setColor(group.devices[d], group.colors[d]);
  }
}

unsigned Region::size() const
{
  return (1 + lastDev - firstDev)* (1 + lastLed - firstLed);
}

// combiner, frequency, frameNumber, params (first call gives frame=0)
typedef boost::function<void (Combiner&, unsigned, unsigned, const std::string&)> AnimationCallback;
class CombinerImpl
{
public:
  CombinerImpl(Region& r)
  : r(r), thread(0) {}
  void run(Combiner* self);
  Region& r;
  Combiner::State* state; // data state
  unsigned char*  leds;  // result
  Combiner::AnimationData animation;
  boost::thread* thread;
  bool running;
  std::string animationName;
  std::string animationParameters;
  unsigned frequency;
  AnimationCallback animationCallback;
};

Combiner::Combiner(Region& r)
{
  _p = new CombinerImpl(r);
  unsigned sz = r.size();
  _p->state = new State[sz];
  memset(_p->state, 0, sizeof(State[sz]));
  _p->leds = new unsigned char[sz*3];
  memset(_p->leds, 0, sz*3);
  for (unsigned i=0; i<6; ++i)
    _p->animation.matrix[i] = new unsigned char[sz*3];
  memset(_p->animation.matrix[0], 0,   sz*3);
  memset(_p->animation.matrix[1], 255, sz*3);
  memset(_p->animation.matrix[2], 85,  sz*3);
  memset(_p->animation.matrix[3], 170, sz*3);
  memset(_p->animation.matrix[4], 255, sz*3);
  _p->running = false;
}

Combiner::~Combiner()
{
  if (_p->running)
    stopAnimation();
  for (unsigned i=0; i<6; ++i)
    delete[] _p->animation.matrix[i];
  delete[] _p->leds;
  delete[] _p->state;
  delete _p;
}

void Combiner::setColor()
{
  for (unsigned i=0, sz = _p->r.size(); i < sz; ++i)
    memcpy(_p->leds + 3*i, _p->animation.matrix[_p->state[i]]+3*i, 3);
  _p->r.setColor(_p->leds);
}


void Combiner::stopAnimation()
{
  if (_p->running && _p->thread)
  {
    _p->running = false;
    _p->thread->interrupt();
    _p->thread->join();
    delete _p->thread;
    _p->thread = 0;
  }
}

void Combiner::setFrequency(unsigned frequency)
{
  _p->frequency = frequency;
}

AnimationCallback acSelf(AnimationCallback cb)
{
  return cb;
}

typedef boost::function<AnimationCallback(const std::string&)> AnimationMaker;
typedef std::map<std::string, std::pair<std::string, AnimationMaker> > AnimationFactory;
AnimationFactory& animationFactory()
{
  static AnimationFactory* af = 0;
  if (!af)
    af = new AnimationFactory();
  return *af;
}
#define REGISTER_ANIMATION(name, desc, func) \
static std::pair<std::string, AnimationMaker> unused_reg##name = (animationFactory()[#name] = std::make_pair(desc, boost::bind(acSelf, func)))
#define REGISTER_ANIMATION_FACTORY(name, desc, func) \
static std::pair<std::string, AnimationMaker> unused_reg##name = (animationFactory()[#name] = std::make_pair(desc, func))

typedef boost::function<Combiner*(Group&, const std::string&, const std::string&)> CombinerMaker;
typedef std::map<std::string, std::pair<std::string, CombinerMaker> > CombinerFactory;
CombinerFactory& combinerFactory()
{
  static CombinerFactory* cf = 0;
  if (!cf)
    cf = new CombinerFactory();
  return *cf;
}

std::vector<Combiner::AnimationInfo>
Combiner::listAnimations()
{
  std::vector<Combiner::AnimationInfo> res;
  for (AnimationFactory::iterator it = animationFactory().begin(); it != animationFactory().end(); ++it)
  {
    res.push_back(std::make_pair(it->first, it->second.first));
  }
  return res;
}

#define REGISTER_COMBINER(name, desc) \
static std::pair<std::string, CombinerMaker> unesed_reg_ ## name = (combinerFactory()[#name] = std::make_pair(desc, &make ## name))

std::vector<std::pair<std::string, std::string> >
Combiner::listCombiners()
{
  std::vector<std::pair<std::string, std::string> > res;
  for (CombinerFactory::iterator it = combinerFactory().begin(); it != combinerFactory().end(); ++it)
  {
    res.push_back(std::make_pair(it->first, it->second.first));
  }
  return res;
}

Combiner* Combiner::create(Group& g, const std::string& name, const std::string& arg)
{
  CombinerFactory::iterator it = combinerFactory().find(name);
  if (it == combinerFactory().end())
    return 0;
  else
    return it->second.second(g, name, arg);
}



void Combiner::selectAnimation(const std::string& name, unsigned frequency, const std::string& params)
{
  if (_p->running
    && _p->animationName == name
    && _p->animationParameters == params)
  { // Check if this is just a frequency change
    _p->frequency = frequency;
    return;
  }
  if (frequency == 0)
  {
    // singleshot target without changing current
    animationFactory().at(name).second(params)(*this, 0, 0, params);
    return;
  }
  stopAnimation();
  _p->animationCallback = animationFactory().at(name).second(params);
  _p->animationName = name;
  _p->animationParameters = params;
  _p->frequency = frequency;
  _p->running = true;
  _p->thread = new boost::thread(&CombinerImpl::run, _p, this);
}

Combiner::State* Combiner::stateData()
{
  return _p->state;
}

Combiner::AnimationData& Combiner::animationData()
{
  return _p->animation;
}

void CombinerImpl::run(Combiner* self)
{
  unsigned frame = 0;
  while (running)
  {
    animationCallback(*self, frame, frequency, animationParameters);
    if (frequency)
      boost::this_thread::sleep(boost::posix_time::microseconds(1000000 / frequency));
  }
}

void Combiner::setDataBinary(unsigned v)
{
  if (_p->r.size() > sizeof(unsigned)*8)
    throw std::runtime_error("Region too big");
  for (unsigned i=0; i<_p->r.size(); ++i)
    _p->state[i] = ((1<<i) & v)? ON:OFF;
  setColor();
}

void Combiner::setData(const std::string& s)
{
  throw std::runtime_error("Call to unimplemented setData");
}

size_t Combiner::size()
{
  return _p->r.size();
}

size_t Combiner::width()
{
  return 1 + _p->r.lastDev - _p->r.firstDev;
}

Digit::Digit(Group& group, unsigned firstDev, unsigned firstLed,
  bool rotate, bool swapx, bool swapy)
: Region(group, firstDev, firstDev+4, firstLed, firstLed + 4)
, Combiner((Region&)*this)
, rotate(rotate)
, swapx(swapx)
, swapy(swapy)
{
  set(0);
}

static const unsigned digitMask[10] = {
  BOOST_BINARY(
    11111
    10001
    10001
    10001
    11111),
  BOOST_BINARY(
    00110
    00010
    00010
    00010
    00111
    ),
  BOOST_BINARY(
    11111
    00001
    11111
    10000
    11111
    ),
  BOOST_BINARY(
    11111
    00001
    01111
    00001
    11111
    ),
  BOOST_BINARY(
    10001
    10001
    11111
    00001
    00001
    ), // 5
  BOOST_BINARY(
    11111
    10000
    11111
    00001
    11111
    ), //6
  BOOST_BINARY(
    11110
    10000
    11111
    10001
    11111
    ), //7
  BOOST_BINARY(
    01111
    00001
    00010
    00100
    01000
    ), //8
    BOOST_BINARY(
    11111
    10001
    11111
    10001
    11111
    ), //9
    BOOST_BINARY(
    11111
    10001
    11111
    00001
    01111
    )
};
void Digit::set(unsigned value)
{
  if (value >9)
    throw std::runtime_error("Digit value must be between 0 and 9");
  size_t v = digitMask[value];
  size_t res = 0;
  for (unsigned i=0; i<25; ++i)
    if (v & (1<<i))
    res += 1<<(24-i);
  if (rotate)
  {
    size_t in = res;
    res = 0;
    for (unsigned x=0; x<5; ++x)
      for (unsigned y=0; y<5; ++y)
        if (in & (1 << (5*x+y)))
          res |= 1<<(5*y+x);
  }
  if (swapx)
  {
    size_t in = res;
    res = 0;
    for (unsigned x=0; x<5; ++x)
      for (unsigned y=0; y<5; ++y)
        if (in & (1 << (5*(4-x)+y)))
          res |= 1<<(5*x+y);
  }
  if (swapy)
  {
    size_t in = res;
    res = 0;
    for (unsigned x=0; x<5; ++x)
      for (unsigned y=0; y<5; ++y)
        if (in & (1 << (5*x+4-y)))
          res |= 1<<(5*x+y);
  }
  setDataBinary(res);
}

void Digit::setData(const std::string& v)
{
  set(strtol(v.c_str(), 0, 0));
}

Combiner* makeDigit(Group& g, const std::string& name, const std::string& args)
{
  std::stringstream s(args);
  unsigned firstDev = 0, lastDev = 0;
  bool a=false, b=false, c=false;
  s >> firstDev >> lastDev >> a >> b >> c;
  return new Digit(g, firstDev, lastDev, a, b, c);
}

REGISTER_COMBINER(Digit, "(firstDev firstLed rotate swapx swapy) One digit from 0 to 9");

class Unary: public Region, public Combiner
{
public:
  Unary(Group& g, unsigned d1, unsigned d2, unsigned l1, unsigned l2);
  virtual void setData(const std::string& d);
  void set(unsigned v);
};

Unary::Unary(Group& g, unsigned d1, unsigned d2, unsigned l1, unsigned l2)
: Region(g, d1, d2, l1, l2)
, Combiner((Region&)*this)
{
}

void Unary::setData(const std::string& d)
{
  set(strtol(d.c_str(), 0, 0));
}

void Unary::set(unsigned v)
{
  if (v > Combiner::size())
    v = Combiner::size();
  State* sptr = stateData();
  for (unsigned i=0; i<v; ++i)
    sptr[i] = ON;
  for (unsigned i=v; i< Region::size(); ++i)
    sptr[i] = OFF;
  Combiner::setColor();
}

Combiner* makeUnary(Group& g, const std::string& name, const std::string& args)
{
  std::stringstream ss(args);
  unsigned p1, p2, p3, p4;
  ss >> p1 >> p2 >> p3 >> p4;
  return new Unary(g, p1, p2, p3, p4);
}
REGISTER_COMBINER(Unary, "(firstDev lastDev firstLed lastLed) Simple unary counter");




struct Coord
{
  Coord()
  :x(0), y(0) {}
  Coord(short x, short y)
  : x(x), y(y) {}
  Coord operator * (const Coord& b) const
  {
    return Coord(x*b.x, y*b.y);
  }
  Coord operator + (const Coord& b) const
  {
    return Coord(x + b.x, y+b.y);
  }
  short lin(short w)
  {
    return x + y*w;
  }
  bool valid(int w, int h)
  {
    return x>=0 && x<w && y>=0 && y<h;
  }
  union
  {
    short v[2];
    struct
    {
      short x;
      short y;
    };
  };
};

void parse(const std::string& v, Coord& target)
{
  size_t p = v.find_first_of(',');
  if (p == v.npos)
  {
    std::cerr << "Coord parse error " << v << std::endl;
  }
  target.x = strtol(v.substr(0, p).c_str(), 0, 0);
  target.y = strtol(v.substr(p+1) .c_str(), 0, 0);
}

void animColor(Combiner& c, unsigned frame, unsigned freq, const std::string& params)
{
  int w = c.width();
  Combiner::AnimationData& ad = c.animationData();
  std::stringstream s(params);
  std::string color, target;
  s >> target;
  std::vector<int> targets;
  for (unsigned p=0; p<target.size(); ++p)
  {
    unsigned tgt = strtol(target.substr(p, 1).c_str(), 0, 0);
    if (tgt > 4)
    {
      std::cerr << "invalid target " << target << std::endl;
      continue;
    }
    targets.push_back(tgt);
  }
  std::vector<DeviceColor> dcs;
  while (true)
  {
    color = "";
    s >> color;
    if (color.empty())
      break;
    size_t p = color.find_first_of(':');
    if (p != color.npos)
    {
      DeviceColor dc(color.substr(p+1), 255);
      Coord c;
      parse(color.substr(0, p), c);
      for (unsigned p=0; p<targets.size(); ++p)
        memcpy(ad.matrix[targets[p]] + c.lin(w), dc.colors, 3);
    }
    else
    {
      DeviceColor dc(color, 255);
      dcs.push_back(dc);
    }
  }
  if (dcs.empty())
    return;
  unsigned sz = c.size();
  for (unsigned p=0; p<targets.size(); ++p)
  {
    unsigned tgt = targets[p];
    for (unsigned i=0; i<sz; ++i)
      memcpy(ad.matrix[tgt] + i*3, dcs[i%dcs.size()].colors, 3);
  }
  c.setColor();
}

struct WaveModel
{
  const char* name;
  int initDir[3];
  unsigned char minV[3];
  unsigned char maxV[3];
  char matrix[9];
  void (*dirChange)(WaveModel& w, int* dir);
  int target;
};

// 1 0 0   -1 0 0   0 1 0   0 -1 0 ...
void dc_alternate(WaveModel& w, int* dir)
{
  if (dir[0] + dir[1] + dir[2] > 0)
  {
    dir[0] *= -1;
    dir[1] *= -1;
    dir[2] *= -1;
  }
  else
  {
    int old[3] = {dir[0], dir[1], dir[2]};
    dir[0] = -old[2];
    dir[1] = -old[0];
    dir[2] = -old[1];
  }
}

// 1 1 0  -1 -1 0  1 0 1  -1 0 -1 ...
void dc_alternate2(WaveModel& w, int* dir)
{
  if (dir[0] + dir[1] + dir[2] > 0)
  {
    dir[0] *= -1;
    dir[1] *= -1;
    dir[2] *= -1;
  }
  else
  {
    int old[3] = {dir[0], dir[1], dir[2]};
    dir[0] = (- old[1] - old[2])/2;
    dir[1] = (- old[2] - old[0])/2;
    dir[2] = (- old[0] - old[1])/2;
  }
}

struct DirChangeFactory
{
  const char* name;
  void (*dirChange)(WaveModel& w, int* dir);
};
DirChangeFactory dirChanges[] =
{
  {"alternate", dc_alternate},
  {"alternate2", dc_alternate},
  {0}
};

#define WM_ALLRANGE {0,0,0}, {255, 255, 255}
#define WMM_REVDIR {-1,0,0, 0,-1,0, 0,0,-1} // reverse direction, no cross-chan
#define WMM_CYCLE {0,1,0, 0,0,1, 1,0,0} // cycle directions between colors
WaveModel waveModels[] = {
  {"waveRGB",    {1, 0, -1}, WM_ALLRANGE, WMM_CYCLE},
  {"alternate",  {1, 0, 0},  WM_ALLRANGE, WMM_CYCLE, &dc_alternate},
  {"revRGB",  {1, 1, 1},     WM_ALLRANGE, WMM_REVDIR},
  {"revR",  {1, 0, 0},       WM_ALLRANGE, WMM_REVDIR},
  {"revG",  {0, 1, 0},       WM_ALLRANGE, WMM_REVDIR},
  {"revB",  {0, 0, 1},       WM_ALLRANGE, WMM_REVDIR},
  {0}
};

void parse(const std::string& val, void (*&dirChange)(WaveModel& w, int* dir))
{
  unsigned i;
  for (i=0; dirChanges[i].name; ++i)
    if (val == dirChanges[i].name)
    {
      dirChange = dirChanges[i].dirChange;
      return;
    }
  unsigned v = strtol(val.c_str(), 0, 0);
  if (v < i)
    dirChange = dirChanges[v].dirChange;
}

template<typename T>
void parse(const std::string& val, T* target, unsigned len)
{
  size_t pPrev = 0;
  while (len)
  {
    size_t p = val.find_first_of(',', pPrev);
    if (p == val.npos)
    {
      *target = strtol(val.substr(pPrev).c_str(), 0, 16);
      --len;
      pPrev = p + 1;
      break;
    }
    else
      *target = strtol(val.substr(pPrev, p - pPrev).c_str(), 0, 16);
    ++target;
    --len;
    pPrev = p + 1;
  }
  if (len || pPrev)
    std::cerr << "List parse error " << val << std::endl;
}

void parseColor(const std::string& val, unsigned char* target)
{
  DeviceColor dc(val, 255);
  memcpy(target, dc.colors, 3);
}

struct Wave
{
  inline bool overflow(int i, int d, unsigned char v)
  {
    int nv = (int)d + (int)v;
    return nv < (int)model.minV[i] || nv > (int)model.maxV[i];
  }
  inline bool overflow(int* d, unsigned char* v)
  {
    return overflow(0, d[0], v[0])
    || overflow(1, d[1], v[1])
    || overflow(2, d[2], v[2]);
  }
  void updateDir(int* d)
  {
    if (model.dirChange)
    {
      (*model.dirChange)(model, d);
      return;
    }
    int d2[3] = {0,0,0};
    for (int i = 0; i<3; ++i)
      for (int j=0; j<3; ++j)
      d2[i] += d[j]*model.matrix[i*3 + j];
    memcpy(d, d2, sizeof(d2));
  }
  Wave()
  : dir(0)
  , stuck(false)
  {
  }
  ~Wave()
  {
    delete[] dir;
  }
  void init(const std::string& params, unsigned sz)
  {
    dir = new int[sz*3];
    model = waveModels[0];
    model.target = 1;
    for (unsigned i=0; i<sz; ++i)
    {
      memcpy(dir + i*3, model.initDir, sizeof(model.initDir));
    }
    std::stringstream s(params);
    while (true)
    {
       std::string p;
       s >> p;
       if (p.empty())
         break;
       size_t sep = p.find_first_of('=');
       if (sep == p.npos)
       {
         p = "model=" + p;
       }
       std::string k = p.substr(0, sep);
       std::string v = p.substr(sep+1);
       if (k == "model")
       {
         // assume model name/index
         bool found = false;
         unsigned i;
         for (i=0; waveModels[i].name; ++i)
         {
           if (v == waveModels[i].name)
           {
             model = waveModels[i];
             model.target = 1;
             found = true;
             break;
           }
         }
         if (!found)
         {
           unsigned match = strtol(v.c_str(), 0, 0);
           if (match < i)
             model = waveModels[match];
           else
             std::cerr << "Wrong model " << p << std::endl;
         }
       }
       else if (k == "initDir") parse(v, model.initDir, 3);
       else if (k == "minV")    parseColor(v, model.minV);
       else if (k == "maxV")    parseColor(v, model.maxV);
       else if (k == "matrix")  parse(v, model.matrix, 9);
       else if (k == "dirChange") parse(v, model.dirChange);
       else if (k == "target") model.target=strtol(v.c_str(), 0, 0);
       else
         std::cerr << "Invalid key " << k << std::endl;
    }
    for (unsigned i=0; i<sz; ++i)
    {
      memcpy(dir + i*3, model.initDir, sizeof(model.initDir));
    }
  }
  void operator()(Combiner& c, unsigned frame, unsigned freq, const std::string& params)
  {

    Combiner::AnimationData& ad = c.animationData();
    unsigned sz = c.size();
    if (!dir)
      init(params, sz);
    unsigned char* m = ad.matrix[model.target];

    for (unsigned p=0; p<sz; ++p)
    {
      unsigned char* pix = m + p*3;
      int* d = dir + p*3;
      for (unsigned r=0; r<3; ++r)
        if (overflow(d, pix))
        updateDir(d);
      if (overflow(d, pix))
      {
        if (!stuck)
          std::cerr << "Animation stuck! " << model.target << " " << p << "   "
            << (int)d[0] << ' '
            << (int)d[1] << ' '
            << (int)d[2] << ' '
            << (int)pix[0] << ' '
            << (int)pix[1] << ' '
            << (int)pix[2] << ' '
            << std::endl;
        stuck = true;
        continue;
      }
      pix[0] += d[0];
      pix[1] += d[1];
      pix[2] += d[2];
    }
    c.setColor();
  }
  int* dir;
  WaveModel model;
  bool stuck;
};

struct AnimPropagate
{
  AnimPropagate() : map(0), m(0) {}
  ~AnimPropagate()
  {
    delete[] map;
    delete[] m;
  }
  AnimPropagate(const AnimPropagate& b)
  : map(0), m(0)
  {
  }
  void operator = (const AnimPropagate& b)
  {
    //nothing
  }
  void init(Combiner& c, const std::string& params)
  {
    unsigned sz = c.size();
    map = new short[sz];
    m = new unsigned char[sz*3];
    int w = c.width();
    int h = sz / w;
    Coord startCorner = Coord(0, 0);
    Coord startDir = Coord(1, 0);
    std::string pattern = "s";
    bool loop = true;
    target = 1;
    std::stringstream s(params);
    while (true)
    {
      std::string kv;
      s >> kv;
      if (kv.empty())
        break;
      size_t p = kv.find_first_of('=');
      if (p == kv.npos)
      {
        std::cerr << "Malformed argument " << kv << std::endl;
        continue;
      }
      std::string k = kv.substr(0, p);
      std::string v = kv.substr(p+1);
      if (k == "pos")
        parse(v, startCorner);
      else if (k == "dir")
        parse(v, startDir);
      else if (k == "pattern")
        pattern = v;
      else if (k == "loop")
        loop = strtol(v.c_str(), 0, 0);
      else if (k == "target")
        target = strtol(v.c_str(), 0, 0);
      else
        std::cerr << "Unexpected argument " << k << std::endl;
    }
    memset(map, 0, sz*sizeof(short));
    // pattern macros
    if (pattern == "left") { pattern = "line"; startDir = Coord(-1, 0);}
    if (pattern == "right"){ pattern = "line"; startDir = Coord(1, 0);}
    if (pattern == "up")   { pattern = "line"; startDir = Coord(0, -1);}
    if (pattern == "down") { pattern = "line"; startDir = Coord(0, 1);}
    #define SETALL(v) for (unsigned i=0; i<sz; ++i) map[i] = v
    // apply pattern
    if (pattern == "line")
      SETALL(startDir.lin(w));
    else if (pattern == "s")
    {
      Coord secondDir(abs(startDir.y), abs(startDir.x));
      secondDir.x *= 1 - 2 * startCorner.x;
      secondDir.y *= 1 - 2 * startCorner.y;
      // We do not know value to set for firstpos. It will be the
      // last coord we encounter.
      int prevc = 0;
      if (!loop)
        prevc = -30000;
      Coord c = startCorner * Coord(h-1, w-1);
      Coord lastValid;
      while (c.valid(h, w))
      {
        lastValid = c;
        int cur = c.lin(h);
        map[cur] = prevc - cur;
        prevc = cur;
        // try going in start direction
        Coord next = c + startDir;
        if (next.valid(h, w))
          c = next;
        else
        {
          // go orthogonal dir and reverse start direction
          c = c + secondDir;
          startDir = startDir * Coord(-1, -1);
        }
      }
      // now set start value to lastValid
      int start = (startCorner * Coord(h-1, w-1)).lin(h);
      map[start] = lastValid.lin(h) - start;
    }
    else
      std::cerr << "Unknown pattern " << pattern << std::endl;
    for (unsigned y=0; y<w; ++y)
    {
      for (unsigned x=0; x<h; ++x)
        std::cerr << map[x+y*h] << ' ';
      std::cerr << std::endl;
    }
  }
  void operator()(Combiner& c, unsigned frame, unsigned freq, const std::string& params)
  {
    if (!map)
      init(c, params);
    unsigned sz = c.size();
    Combiner::AnimationData& ad = c.animationData();
    unsigned char* mdst = ad.matrix[target];
    memcpy(m, mdst, sz*3);
    for (unsigned i=0; i<sz; ++i)
    {
      short target = i + map[i];
      if (target < 0 || target >= (int)sz)
        memset(mdst+i*3, 0, 3);
      else
        memcpy(mdst+i*3, m+target*3, 3);
    }
    c.setColor();
  }
  short* map; // map[i] = relative coordinate of cell to fetch data
  unsigned char* m;
  unsigned target;
};

AnimationCallback makeWave(const std::string& p)
{
  return Wave();
}
AnimationCallback makePropagate(const std::string& p)
{
  return AnimPropagate();
}

REGISTER_ANIMATION(Color, "(LAYER COLOR...)Just set a color", animColor);
REGISTER_ANIMATION_FACTORY(Wave, "(model=MODOEL target=LAYER) A nice wave", makeWave);
REGISTER_ANIMATION_FACTORY(Propagate, "(pattern=MODEL target=LAYER dir=x,y pos=x,y) propagation", makePropagate);

}
