/* Copyright (C) 2012,2013, Matthieu Nottale
* See README.md for license
*/

#ifndef LIBCLIENT_HH
#define LIBCLIENT_HH
#include <string>
#include <vector>

namespace ptrim
{
  union RGB
  {
    struct {
      unsigned char r,g, b;
    };
    unsigned char v[3];
  };

  class DeviceColor
  {
  public:
    DeviceColor();
    DeviceColor(const std::string& data, unsigned char brigthness);
    void set(unsigned char c);
    void set(RGB c);
    void set(unsigned led, RGB c);
    std::string toString() const;
    RGB colors[8];
    unsigned char brightness;
  };


  class Client
  {
  public:
    virtual ~Client() {}
    virtual const std::vector<std::string>& devices() = 0;
    virtual void setColor(const std::string& device, const DeviceColor& color) = 0;
  };

  class NetClientImpl;
  class NetClient: public Client
  {
  public:
    NetClient();
    ~NetClient();
    void connect(const std::string& host, const std::string& port);
    virtual const std::vector<std::string>& devices();
    virtual void setColor(const std::string& device, const DeviceColor& color);
    NetClientImpl* _p;
  };
  class Mask
  {
  public:
    unsigned char mask[32];
  };
  /// Group of devices, with support for masked set
  class Group
  {
  public:
    Group(Client& client, const std::vector<std::string> devices);
    void setColor(const std::vector<DeviceColor>& data, const Mask& m);
    Client& client;
    std::vector<std::string> devices;
    std::vector<DeviceColor> colors;
  };
  /// Rectangular mask on a set of devices. Use Group backend.
  class Region
  {
  public:
    Region(Group& g, unsigned firstDev, unsigned lastDev, unsigned firstLed, unsigned lastLed);
    void setColor(const unsigned char* data); //length nDev * nLeds * 3
    unsigned size() const;
    Group& group;
    unsigned firstDev, lastDev;
    unsigned firstLed, lastLed;
  };

  class CombinerImpl;
  /** Combine multiple sources on the same output.
   *
   * The idea is to have a data source that set a state per led,
   * and an animation source that varies the output within the
   * constraints set by the data source.
   *
   * Data source can be one bit per led (on, off)
   * or two bits (off, on_1 on_2 on_3)
   *
   * Animation source provides and updates a
   * per-state array of 3 (rgb) per-led value in 0-255.
   *
  */
  class Combiner
  {
  public:
    struct AnimationData
    {
      unsigned char* matrix[5];
    };
    enum State
    {
      OFF = 0,  // of at any intensity
      ON  = 1,  // on at any intensity
      ON_1 = 2, // on at intensity 1 (one color)
      ON_2 = 3, // on at intensity 2 (two colors)
      ON_3 = 4, // on at intensity 3 (three colors)
    };
    Combiner(Region& r);
    ~Combiner();
    // Implementation-dependant data setter
    virtual void setData(const std::string& s);
    /// set data, one bit per led, region size must be < sizeof(unsigned)*8
    void setDataBinary(unsigned binData);

    /// name, description
    typedef std::pair<std::string, std::string> AnimationInfo;

    /// Return the list of known animations
    static std::vector<AnimationInfo> listAnimations();
    static std::vector<std::pair<std::string, std::string> > listCombiners();
    static Combiner* create(Group& g, const std::string& factory, const std::string& args);
   
    /// Set and start animation pattern.
    void selectAnimation(const std::string& name, unsigned frequency, const std::string& params);
    void stopAnimation();
    void setFrequency(unsigned frequency);

    // Combine state and animation data, and apply new colors
    void setColor();

    State*         stateData();
    AnimationData& animationData();
    size_t size();
    size_t width();

    CombinerImpl* _p;
  };
  /// Display one digit on an 5x5 region
  class Digit: public Region, public Combiner
  {
  public:
    Digit(Group& group, unsigned firstDev, unsigned firstLed, bool rotate=false, bool swapx=false, bool swapy=false);
    /// Set digit value from 0 to 9
    void set(unsigned value);
    virtual void setData(const std::string& s);
  protected:
    bool rotate, swapx, swapy;
  };
}
#endif
