/* Copyright (C) 2012,2013, Matthieu Nottale
* See README.md for license
*/

#include <map>
#include <vector>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <boost/asio.hpp>
#include <boost/thread.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>

#ifndef BACKEND_HIDRAW
#include "ptrim-lib.h"
#else
  char* getSerial(int fd)
  {
    /*fake
    static std::string res;
    res = boost::lexical_cast<std::string>(fd);
    return res.c_str();
    */
    static char cmd[32];
    memset(cmd,0x00,32);
    cmd[0] = 0x09;
    write(fd, cmd, 32);
    int res = read(fd, cmd, 32);
    if (res != 32)
    {
      std::cerr  << "read failure " << res << ' ' << strerror(res) << std::endl;
    }
    memmove(cmd, cmd+1, 31);
    int p = 0;
    for ( p = strlen(cmd); p>0; --p)
      if (cmd[p] == ' ') break;
    return cmd+p+1;
  }
  void setColor(int fd, unsigned char* colorData, unsigned char brightness)
  {
    unsigned char cmd[32];
    memset(cmd, 0, 32);
    cmd[0] = 0;
    memcpy(cmd+1, colorData, 24);
    cmd[25] = brightness;
    int res = write(fd, cmd, 32);
    if (res != 32)
      std::cerr <<"write failure " << res << " " << strerror(errno) << std::endl;
    res = read(fd, cmd, 32);
    if (res != 32)
      std::cerr <<"read failure " << res << " " << strerror(errno) << std::endl;
  }
  void stop(int fd)
  {
    unsigned char cmd[32];
    memset(cmd,0x00,32);
    cmd[0] = 0x03;
    int res = write(fd, cmd, 32);
    if (res != 32)
      std::cerr <<"write failure " << res << " " << strerror(errno) << std::endl;
  }
  void start(int fd)
  {
    unsigned char cmd[32];
    memset(cmd,0x00,32);
    cmd[0] = 0x02;
    int res = write(fd, cmd, 32);
    if (res != 32)
      std::cerr <<"write failure " << res << " " << strerror(res) << std::endl;
  }
#endif
#include "libclient.hh"

static bool debugOn = getenv("DEBUG");

#define DEBUG(a) do { if (debugOn) std::cerr << a << std::endl;} while(false)


using namespace boost::asio;
using boost::system::error_code;

class Device
{
public:
#ifndef BACKEND_HIDRAW
  hid_device*
#else
  int
#endif
              handle;
  std::string serial;
};

typedef std::map<std::string, Device> Handles;
Handles handles;

typedef std::map<std::string, ptrim::Combiner*> Widgets;
Widgets widgets;

boost::asio::io_service io;
ip::tcp::acceptor* acceptor = 0;

class ServerClient: public ptrim::Client
{
public:
  virtual const std::vector<std::string>& devices();
  virtual void setColor(const std::string& device, const ptrim::DeviceColor& color);
};

class NullClient: public ptrim::Client
{
public:
  virtual const std::vector<std::string>& devices() { return _devices;}
  virtual void setColor(const std::string& device, const ptrim::DeviceColor& color) {}
  std::vector<std::string> _devices;
};

ptrim::Client* mainClient;
ptrim::Group* ptrimGroup;

std::vector<std::string> listDevices()
{
  std::vector<std::string> res;
#ifndef BACKEND_HIDRAW
  hid_device_info* devs = hid_enumerate(0x26f3, 0x1000);
  hid_device_info* cur_dev = devs;	
  while (cur_dev) {
    res.push_back(cur_dev->path);
    cur_dev = cur_dev->next;
  }
  hid_free_enumeration(devs);
#else
  for (unsigned i=0; i<10; ++i)
  {
    std::stringstream s;
    s << "/dev/hidraw" << i;
    struct stat st;
    int statres = stat(s.str().c_str(), &st);
    if (!statres)
      res.push_back(s.str());
  }
#endif
  return res;
}


void close()
{
  DEBUG("Closing handles...");
  for (Handles::iterator it = handles.begin(); it!= handles.end(); ++it)
  {
#ifndef BACKEND_HIDRAW
    hid_close(it->second.handle);
#else
    close(it->second.handle);
#endif
  }
  handles.clear();
}





void refreshDevices()
{
  DEBUG("Getting device list...");
  std::vector<std::string> available = listDevices();
  DEBUG("Done " << available.size());
  // remove dead
  for (Handles::iterator it = handles.begin(); it != handles.end(); )
  {
    Handles::iterator it2 = it;
    ++it2;
    if (std::find(available.begin(), available.end(), it->first) == available.end())
    {
      #ifndef BACKEND_HIDRAW
      hid_close(it->second.handle);
      #else
      close(it->second.handle);
      #endif
      handles.erase(it);
    }
    it = it2;
  }
  // add new
  for (unsigned i=0; i<available.size(); ++i)
  {
    if (handles.find(available[i]) == handles.end())
    {
#ifndef BACKEND_HIDRAW
      DEBUG("Opening device " << available[i]);
      hid_device* h = hid_open_path(available[i].c_str());
      DEBUG("Done " << h);
      if (!h)
      {
        std::cerr << "Could not open " << available[i] << std::endl;
        continue;
      }
#else
      int h = open(available[i].c_str(), O_RDWR);
      if (h < 0)
        {
        std::cerr << "Could not open " << available[i] << std::endl;
        continue;
      }
#endif
      /* Under the raspberry pi the damm thing keep freezing. Delay help.
      */
      usleep(100000);
      Device d;
      d.handle = h;
      DEBUG("Fetching serial");
      d.serial = getSerial(h);
      DEBUG("Got " << d.serial);
      handles[available[i]] = d;
      usleep(100000);
    }
  }
}

template<typename F>
void applyAll(std::stringstream& s, F f)
{
  while (true)
  {
    std::string serial;
    s >> serial;
    if (serial.empty())
      break;
    bool found = false;
    for (Handles::iterator it = handles.begin(); it != handles.end(); ++it)
    {
      if (it->second.serial == serial)
      {
        f(it->second.handle);
        found = true;
        break;
      }
    }
    if (!found)
    {
      unsigned v = strtol(serial.c_str(), 0, 0);
      if (ptrimGroup->devices.size() > v)
      {
        serial = ptrimGroup->devices[v];
        for (Handles::iterator it = handles.begin(); it != handles.end(); ++it)
        {
          if (it->second.serial == serial)
          {
            f(it->second.handle);
            found = true;
            break;
          }
        }
      }
    }
    if (!found)
      std::cerr << "Device not found: " << serial;
  }
}

void processCommand(const std::string& command, ip::tcp::socket* so)
{
  std::stringstream s(command);
  std::string arg0;
  s >> arg0;
  if (arg0.empty())
   return;
  if (arg0[0] == 'h')
  {
    std::string reply =
    "devices: space-separated device list\n"
    "stop DEVICES...: stop embeded animation\n"
    "start DEVICES..: start embeded animation\n"
    "color COL BRIGHTNESS DEVICES...: set color(3,6,24,48 bytes: hex values)\n"
    "anims: list available animations\n"
    "wlist: list available widgets\n"
    "wcreate UID NAME ARGS: create a new widget\n"
    "wset UID VAL: set widget displayed value\n"
    "wanim UID ANIM FREQ ANIMARGS: assign animation to widget\n"
    "wfreq UID FREQ: change current animation frequency\n"
    "wstop UID: stop animation\n"
    "dump: dump color state of all devices\n"
    "reorder DEVICES...: Update ordering of devices\n"
    "\n";
     write(*so, buffer(reply.c_str(), reply.length()));
  }
  else if (arg0 == "devices")
  {
    std::string reply;
    std::vector<std::string> d = ptrimGroup->devices;
    for (unsigned i=0; i<d.size(); ++i)
      reply += d[i] + "\n";
    reply += "\n";
    write(*so, buffer(reply.c_str(), reply.length()));
  }
  else if (arg0 == "reorder")
  {
    std::vector<std::string> nd;
    while (s.good())
    {
      std::string d;
      s >> d;
      if (!d.empty())
        nd.push_back(d);
    }
    ptrimGroup->devices = nd;
  }
  else if (arg0 == "stop")
  {
    applyAll(s, &stop);
  }
  else if (arg0 == "start")
  {
    applyAll(s, &start);
  }
  else if (arg0 == "color")
  {
    std::string col;
    unsigned brightness;
    s >> col >> brightness;
    try
    {
      ptrim::DeviceColor dc(col, brightness);
      std::cerr << dc.toString() << std::endl;
      applyAll(s, boost::bind(&setColor, _1, (unsigned char*)dc.colors, brightness));
    }
    catch(const std::exception& e)
    {
      std::cerr << "Invalid color: " << e.what() << std::endl;
    }
  }
  else if (arg0 == "dump")
  {
    std::string res;
    for (unsigned i=0; i< ptrimGroup->colors.size(); ++i)
      res += ptrimGroup->colors[i].toString() + "\n";
    res += "\n";
    write(*so, buffer(res.c_str(), res.size()));
  }
  else if (arg0 == "anims")
  {
    std::vector<ptrim::Combiner::AnimationInfo> anims = ptrim::Combiner::listAnimations();
    std::stringstream res;
    for (unsigned i=0; i<anims.size(); ++i)
    {
      res << anims[i].first <<" " << anims[i].second << "\n";
    }
    res << "\n";
    std::string s = res.str();
    write(*so, buffer(s.c_str(), s.size()));
  }
  else if (arg0 == "wlist")
  {
    std::vector<std::pair<std::string, std::string> > list = ptrim::Combiner::listCombiners();
    std::stringstream res;
    for (unsigned i=0; i<list.size(); ++i)
      res << list[i].first << " " << list[i].second << "\n";
    res << "\n";
    std::string s = res.str();
    write(*so, buffer(s.c_str(), s.size()));
  }
  else if (arg0[0] == 'w')
  {
    std::string uid;
    s >> uid;
    if (arg0 == "wcreate")
    {
      std::string factory, args;
      s >> factory;
      std::getline(s, args);
      ptrim::Combiner* c = ptrim::Combiner::create(*ptrimGroup, factory, args);
      if (!c)
        std::cerr << "Widget creation failed";
      else
        widgets[uid] = c;
      return;
    }
    Widgets::iterator it = widgets.find(uid);
    if (it == widgets.end())
    {
      std::cerr << "No such widget " << uid << std::endl;
      return;
    }
    else if (arg0 == "wset")
    {
      std::string  value;
      s  >> value;
      it->second->setData(value);
    }
    else if (arg0 == "wanim")
    {
      std::string anim, animParam;
      unsigned frequency;
      s >> anim >> frequency;
      std::getline(s, animParam);
      try
      {
        it->second->selectAnimation(anim, frequency, animParam);
      }
      catch(const std::exception& e)
      {
        std::cerr << "Error setting animation: " << e.what();
      }
    }
    else if (arg0 == "wfreq")
    {
      unsigned frequency;
      s >> frequency;
      it->second->setFrequency(frequency);
    }
    else if (arg0 == "wstop")
    {
      it->second->stopAnimation();
    }
  }
  else
  {
    std::cerr << "invalid command " << arg0 << std::endl;
  }
}
void readMsg(ip::tcp::socket* s, streambuf* response, size_t len, const error_code& error);

void readMsg(ip::tcp::socket* s, streambuf* response, size_t len, const error_code& error)
{
  if (error)
  {
    std::cerr << "connection lost: " << error.message() << std::endl;
    delete response;
    delete s;
    return;
  }
  std::istream response_stream(response);
  std::string line;
  std::getline(response_stream, line);
  processCommand(line, s);
  async_read_until(*s, *response, "\n", boost::bind(&readMsg, s, response, _2, _1));
}

void acceptOne(ip::tcp::socket* s, const error_code& error)
{
  if (error)
  {
    std::cerr << "accept error " << error.message() << std::endl;
    delete s;
    return;
  }
  streambuf* response = new streambuf();
  async_read_until(*s, *response, "\n", boost::bind(&readMsg, s, response, _2, _1));
  s = new ip::tcp::socket(io);
  acceptor->async_accept(*s, boost::bind(&acceptOne, s, _1));
}

int startListener(int port)
{
  ip::tcp::endpoint endpoint(ip::tcp::v4(), port);
  acceptor = new ip::tcp::acceptor(io, endpoint);
  port = acceptor->local_endpoint().port();
  ip::tcp::socket* s = new ip::tcp::socket(io);
  acceptor->async_accept(*s, boost::bind(&acceptOne, s, _1));
  return port;
}

const std::vector<std::string>& ServerClient::devices()
{
  static std::vector<std::string> reply;
  reply.clear();
  for (Handles::iterator it = handles.begin(); it != handles.end(); ++it)
    reply.push_back(it->second.serial);
  return reply;
}

void ServerClient::setColor(const std::string& device, const ptrim::DeviceColor& color)
{
  DEBUG("client::setColor " << device << ' ' << color.toString());
  for (Handles::iterator it = handles.begin(); it != handles.end(); ++it)
  {
    if (it->second.serial == device)
    {
      ::setColor(it->second.handle, (unsigned char*)color.colors, 255);
      return;
    }
  }
  throw std::runtime_error("No such device");
}


// Portable console asynchronous IO is complicated.
// So don't bother, use two threads
void writeStdOut(ip::tcp::socket* s)
{
  streambuf data;
  while (true)
  {
    read_until(*s, data, "\n");
    std::istream response_stream(&data);
    std::string line;
    std::getline(response_stream, line);
    std::cout << line << std::endl;
  }
}
void readStdin(int port)
{
  io_service ioStdin;
  ip::tcp::socket* sStdin = new ip::tcp::socket(ioStdin);
  ip::tcp::endpoint ep(ip::address::from_string("127.0.0.1"), port);
  sStdin->connect(ep);
  boost::thread t(writeStdOut, sStdin);
  while (true)
  {
    std::string line;
    std::getline(std::cin, line);
    line += '\n';
    write(*sStdin, buffer(line.c_str(), line.size()));
  }
}

void processFiles(int port, std::vector<std::string> files)
{
  io_service io;
  ip::tcp::socket* s = new ip::tcp::socket(io);
  ip::tcp::endpoint ep(ip::address::from_string("127.0.0.1"), port);
  s->connect(ep);
  for (unsigned i=0; i<files.size(); ++i)
  {
    std::ifstream file (files[i].c_str());
    char buf[512];
    while (file.read(buf, sizeof(buf)).gcount() > 0)
        write(*s, buffer(buf, file.gcount()));
  }
}

void onSignal(int)
{
  io.stop();
}

int main(int argc, char** argv)
{
  DEBUG("init hid");
  int fail = hid_init();
  if (fail)
  {
    std::cerr << "hid_init failure";
    return 1;
  }
  int port = 0;
  int fake = -1;
  int p = 1;
  bool listenCin  = true;
  std::vector<std::string> files;
  while (p < argc)
  {
    std::string arg = argv[p++];
    if (arg == "-f")
    {
      fake = strtol(argv[p++], 0, 0);
    }
    if (arg == "-n")
    {
      listenCin = false;
    }
    else if (arg == "-h" || arg == "--help")
    {
      std::cerr << "usage: " << argv[0] << " [-n] [-f fakeDeviceCount] [-p port] [FILES...]" << std::endl;
      std::cerr << "\t -n: Do not listen on stdin" << std::endl;
      std::cerr << "\t -p PORT: TCP port to listen on" << std::endl;
      std::cerr << "\t FILES: files containing commands to run" << std::endl;
      return 0;
    }
    else if (arg == "-h" || arg == "--port")
      port = strtol(argv[p++], 0, 0);
    else
      files.push_back(arg);
  }
  if (fake != -1)
  {
    DEBUG("faking " << fake << " devices");
    std::vector<std::string> devices;
    for (int i=0; i<fake; ++i)
      devices.push_back(boost::lexical_cast<std::string>(i));
    mainClient = new NullClient();
    ((NullClient*)mainClient)->_devices = devices;
  }
  else
  {
    refreshDevices();
    mainClient = new ServerClient();

  }
  ptrim::DeviceColor dc;
  ptrimGroup = new ptrim::Group(*mainClient, mainClient->devices());
  port = startListener(port);
  std::cerr << "listening on stdin and tcp " << port << " . Type help for command list." << std::endl;
  if (listenCin)
    boost::thread t(readStdin, port);
  if (!files.empty())
    boost::thread t(processFiles, port, files);
  signal(SIGTERM, onSignal);
  #ifdef __linux__
  signal(SIGQUIT, onSignal);
  #endif
  signal(SIGINT, onSignal);
  io.run();

  close();
}
